/* 
 * Copyright 2013 ScaleOut Software, Inc.
 * 
 * LICENSE AND DISCLAIMER
 * ----------------------
 * This material contains sample programming source code ("Sample Code").
 * ScaleOut Software, Inc. (SSI) grants you a nonexclusive license to compile, 
 * link, run, display, reproduce, and prepare derivative works of 
 * this Sample Code.  The Sample Code has not been thoroughly
 * tested under all conditions.  SSI, therefore, does not guarantee
 * or imply its reliability, serviceability, or function. SSI
 * provides no support services for the Sample Code.
 *
 * All Sample Code contained herein is provided to you "AS IS" without
 * any warranties of any kind. THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGMENT ARE EXPRESSLY
 * DISCLAIMED.  SOME JURISDICTIONS DO NOT ALLOW THE EXCLUSION OF IMPLIED
 * WARRANTIES, SO THE ABOVE EXCLUSIONS MAY NOT APPLY TO YOU.  IN NO 
 * EVENT WILL SSI BE LIABLE TO ANY PARTY FOR ANY DIRECT, INDIRECT, 
 * SPECIAL OR OTHER CONSEQUENTIAL DAMAGES FOR ANY USE OF THE SAMPLE CODE
 * INCLUDING, WITHOUT LIMITATION, ANY LOST PROFITS, BUSINESS 
 * INTERRUPTION, LOSS OF PROGRAMS OR OTHER DATA ON YOUR INFORMATION
 * HANDLING SYSTEM OR OTHERWISE, EVEN IF WE ARE EXPRESSLY ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGES.
 */

package com.scaleoutsoftware.demos.common;

import java.io.Serializable;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.List;


/**
 * Represents an investment strategy consisting of core and hedge positions.
 */
public class Strategy implements Serializable {
    private static final long serialVersionUID = 1L;
    private final int id;
    private final List<Position> positions = new ArrayList<Position>();
    private boolean coreAlerted;
    private boolean hedgeAlerted;

    /**
     * Constructor
     * @param id  strategy identifier
     */
    public Strategy(int id)
    {
        this.id = id;
    }
    
    @Override
    public String toString()
    {
        return Integer.toString(id);
    }

    /**
     * @return strategy identifier
     */
    public int getId() {
        return id;
    }

    /**
     * @return the collection of positions held by this strategy
     */
    public List<Position> getPositions() {
        return positions;
    }

    /**
     * @return whether any of the core positions in this strategy are alerted.
     */
    public boolean isCoreAlerted() {
        return coreAlerted;
    }

    /**
     * @return whether any of the hedge positions in this strategy are alerted.
     */
    public boolean isHedgeAlerted() {
        return hedgeAlerted;
    }

    /**
     * Calulates the total value of the core/hedge positions in this strategy
     * @param type the type of position (core vs. hedge) whose total value is to be calculated
     * @return dollar value of the strategy
     */
    public BigDecimal GetTotalStrategyValue(PositionType type)
    {
        BigDecimal total = BigDecimal.ZERO;
        for (Position position : this.getPositions())
        {
            if (position.getPositionType() == type)
                total = total.add(position.getExposure());
        }

        return total;
    }

    /**
     * Updates each position's actual percent allocation within the strategy. Call after updating
     * prices for positions' securities to refresh holdings' percentages within the strategy. If 
     * a position's actual allocation falls too far outside of the desired allocation then it will
     * be set to an alerted state.
     * 
     * @param positionType           the type of position (core vs. hedge) whose percentages are to be refreshed.
     * @param alertThresholdPercent  the maximum percentage that a position can vary from its targeted value before it enters
     *                               an alerted state.
     */
    public void updateActualAllocations(PositionType positionType, BigDecimal alertThresholdPercent) {
        BigDecimal totalVal = GetTotalStrategyValue(positionType);
        boolean alertStrategy = false;

        for (Position position : this.getPositions())
        {
            if (position.getPositionType() == positionType)
            {
                position.setActualAllocationPercent(position.getExposure().divide(totalVal, RoundingMode.HALF_UP).multiply(BigDecimal.valueOf(100)).setScale(3));
                if (position.getDeviation().abs().subtract(alertThresholdPercent).compareTo(BigDecimal.ZERO) > 0)
                {
                    alertStrategy = true;
                    position.setAlerted(alertStrategy);
                }
                else
                    position.setAlerted(false);
            }
        }

        if (positionType == PositionType.Core)
            coreAlerted = alertStrategy;
        else
            hedgeAlerted = alertStrategy;
    }

}
