/* 
 * Copyright 2013 ScaleOut Software, Inc.
 * 
 * LICENSE AND DISCLAIMER
 * ----------------------
 * This material contains sample programming source code ("Sample Code").
 * ScaleOut Software, Inc. (SSI) grants you a nonexclusive license to compile, 
 * link, run, display, reproduce, and prepare derivative works of 
 * this Sample Code.  The Sample Code has not been thoroughly
 * tested under all conditions.  SSI, therefore, does not guarantee
 * or imply its reliability, serviceability, or function. SSI
 * provides no support services for the Sample Code.
 *
 * All Sample Code contained herein is provided to you "AS IS" without
 * any warranties of any kind. THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGMENT ARE EXPRESSLY
 * DISCLAIMED.  SOME JURISDICTIONS DO NOT ALLOW THE EXCLUSION OF IMPLIED
 * WARRANTIES, SO THE ABOVE EXCLUSIONS MAY NOT APPLY TO YOU.  IN NO 
 * EVENT WILL SSI BE LIABLE TO ANY PARTY FOR ANY DIRECT, INDIRECT, 
 * SPECIAL OR OTHER CONSEQUENTIAL DAMAGES FOR ANY USE OF THE SAMPLE CODE
 * INCLUDING, WITHOUT LIMITATION, ANY LOST PROFITS, BUSINESS 
 * INTERRUPTION, LOSS OF PROGRAMS OR OTHER DATA ON YOUR INFORMATION
 * HANDLING SYSTEM OR OTHERWISE, EVEN IF WE ARE EXPRESSLY ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGES.
 */

package com.scaleoutsoftware.demos.common;

import com.scaleoutsoftware.soss.client.EvalArgs;
import com.scaleoutsoftware.soss.client.EvalObjectDisposition;
import com.scaleoutsoftware.soss.client.Invokable;
import java.util.HashMap;

/**
 * Class used to update prices in a cached Strategy object in parallel across the entire
 * in-memory data grid.
 * 
 * Logic is invoked via ScaleOut's Parallel Method Invocation (PMI) mechanism.
 */
public class StrategyUpdater implements Invokable<Strategy, StrategyUpdateArgs, AlertSummary> {

    /**
     * An "Eval" method that is invoked by ScaleOut's Parallel Method Invocation (PMI) engine to update 
     * prices in a strategy's positions. A NamedCache.invoke() call will cause this method to be invoked in 
     * parallel across the farm for every strategy stored in the in-memory data grid.
     * 
     * @param strategy            A strategy that has been stored in the in-memory data grid.
     * @param strategyUpdateArgs  A PMI "parameter object" that was passed into the NamedCache.invoke() call,
     *                            allowing the parameter to be applied to all positions in the cache by this eval method. 
     *                            The parameter in this case contains a collection of current security prices
     *                            and the user's desired alert threshold.
     * @param strategyEvalArgs    Helper argument that allows further control of the PMI eval operation.
     * @return                    An AlertSummary object containing the alert status of single item 
     *                            (the strategy being evaluated by this eval call). All results returned 
     *                            by invocations of this method are merged back into a single value that 
     *                            is returned to the original NamedCache.invoke caller.
     */
    @Override
    public AlertSummary eval(Strategy strategy, StrategyUpdateArgs strategyUpdateArgs, EvalArgs<Strategy> strategyEvalArgs) //throws InvokeException, InterruptedException
    {
        HashMap<String, Security> latestPrices = strategyUpdateArgs.getLatestPrices();
        
        for (Position position : strategy.getPositions())
        {
            position.setPrice(latestPrices.get(position.getTicker()).getLatestPrice());
        }

        strategy.updateActualAllocations(PositionType.Core, strategyUpdateArgs.getMaxAllowedDeviationPercent());
        strategy.updateActualAllocations(PositionType.Hedge, strategyUpdateArgs.getMaxAllowedDeviationPercent());

        AlertSummary result = new AlertSummary(strategy.getId(), 
                                               (strategy.isCoreAlerted() || strategy.isHedgeAlerted()),
                                               strategy.getPositions().size());
        
        // set the object disposition to *update* here so that our strategy
        // updates are persisted back to ScaleOut in-memory data grid.
        strategyEvalArgs.setDisposition(EvalObjectDisposition.Update);

        return result;
    }

    /**
     *  A "Merge" method that is invoked by ScaleOut's Parallel Method Invocation (PMI) engine to combine
     * results from a distributed operation. In this case, we're combining alert statuses that were
     * triggered by a price update -- the summary that is finally returned by the NamedCache.invoke call will
     * ultimately contain the alert status for every single strategy in the in-memory data grid.
     * @param alerts1  First alert summary.
     * @param alerts2  Second alert summary.
     * @return         Combined alert summary.
     */
    @Override
    public AlertSummary merge(AlertSummary alerts1, AlertSummary alerts2)  //throws InvokeException, InterruptedException
    {
        if (alerts1 == null)
            alerts1 = new AlertSummary();
        
        alerts1.merge(alerts2);
        
        return alerts1;
    }
}
