/* 
 * Copyright 2013 ScaleOut Software, Inc.
 * 
 * LICENSE AND DISCLAIMER
 * ----------------------
 * This material contains sample programming source code ("Sample Code").
 * ScaleOut Software, Inc. (SSI) grants you a nonexclusive license to compile, 
 * link, run, display, reproduce, and prepare derivative works of 
 * this Sample Code.  The Sample Code has not been thoroughly
 * tested under all conditions.  SSI, therefore, does not guarantee
 * or imply its reliability, serviceability, or function. SSI
 * provides no support services for the Sample Code.
 *
 * All Sample Code contained herein is provided to you "AS IS" without
 * any warranties of any kind. THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGMENT ARE EXPRESSLY
 * DISCLAIMED.  SOME JURISDICTIONS DO NOT ALLOW THE EXCLUSION OF IMPLIED
 * WARRANTIES, SO THE ABOVE EXCLUSIONS MAY NOT APPLY TO YOU.  IN NO 
 * EVENT WILL SSI BE LIABLE TO ANY PARTY FOR ANY DIRECT, INDIRECT, 
 * SPECIAL OR OTHER CONSEQUENTIAL DAMAGES FOR ANY USE OF THE SAMPLE CODE
 * INCLUDING, WITHOUT LIMITATION, ANY LOST PROFITS, BUSINESS 
 * INTERRUPTION, LOSS OF PROGRAMS OR OTHER DATA ON YOUR INFORMATION
 * HANDLING SYSTEM OR OTHERWISE, EVEN IF WE ARE EXPRESSLY ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGES.
 */

package com.scaleoutsoftware.demos.marketsimulator;

import com.scaleoutsoftware.soss.client.*;
import com.scaleoutsoftware.demos.common.*;

import java.io.*;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.*;

/**
 * Market simulator application.
 * 
 * Creates simulated securities and adds strategies (consisting of multiple hedge/core positions)
 * to the ScaleOut StateServer in-memory data grid. The app then starts a worker thread that 
 * simulates incoming price updates from a stock exchange.
 * 
 * Once the simulation is started, the program uses ScaleOut's PMI (parallel method invocation)
 * engine to periodically update the large number of investment strategy objects across the farm
 * in parallel with the latest market prices and to evaluate them for alert conditions.
 */
public class Main {
    private static final long MINIMUM_INTERVAL_BETWEEN_PMI_CALLS = 1000; // in milliseconds
    private static final int SYMBOL_COUNT = 3000; // rough estimate of securities in NYSE

    /**
     * Main entry point for the market simulator application.
     * 
     * This method generates simulated securities and investment strategies, then fires off a
     * market simulator thread that continuously updates security prices. ScaleOut's PMI feature
     * is invoked periodically to updated all strategies with latest market prices and generate
     * alerts.
     * 
     * @param args Command line arguments.
     *             Expects: strategyCount, priceUpdatesPerSec, positionsPerStrategy
     */
    public static void main(String[] args) {
        int strategyCount, priceUpdatesPerSec, positionsPerStrategy;

        try {
            if (args.length != 3)
                throw new IllegalArgumentException();

            strategyCount = Integer.parseInt(args[0]);
            priceUpdatesPerSec = Integer.parseInt(args[1]);
            positionsPerStrategy = Integer.parseInt(args[2]);
        } catch (Exception ex) {
            System.out.println("required args: strategyCount priceUpdatesPerSec positionsPerStrategy");
            return;
        }

        System.out.format("%tT: Initializing securities...\n",
                Calendar.getInstance());

        HashMap<String, Security> securities;
        try {
            securities = generateInitialSecurities(SYMBOL_COUNT);
        } catch (Exception ex) {
            System.out.println("Unable to generate securities: " + ex.getMessage());
            return;
        }

        System.out.format("%tT: Loading %d strategies into the data grid...\n",
                Calendar.getInstance(), strategyCount);

        try {
            createStrategies(strategyCount, positionsPerStrategy, securities);
        } catch (Exception ex) {
            System.out.println("Unable to create strategies: " + ex.getMessage());
            return;
        }

        // Configure and load the invocation grid workers that will automatically be started up 
        // on every host in the farm.
        NamedCache strategyCache;
        InvocationGrid grid;
        try {
            System.getProperty("user.dir");
            
            strategyCache = CacheFactory.getCache(Settings.STRATEGY_CACHE_NAME);
            grid = new InvocationGridBuilder("Strategy update worker grid " + UUID.randomUUID().toString()).
                    addJar("./strategyalerts.jar").
                    setJVMParameters("-Xmx1g").
                    setLingerTimeMinutes(5).
                    load(true);

            strategyCache.setInvocationGrid(grid);
        } catch (Exception ex) {
            System.out.println("Unable to Load invocation grid: " + ex.getMessage());
            return;
        }

        // Start up the simulated exchange feed that provides incoming price updates for the 
        // securities that we're interested in:
        ExchangeSimulator exchangeSimulator = new ExchangeSimulator(priceUpdatesPerSec, securities);
        Thread simulatorThread = new Thread(exchangeSimulator);
        simulatorThread.setDaemon(true);
        simulatorThread.start();
        
        // Periodically update all strategies in the in-memory data grid. Continue until the 
        // program is terminated.
        invokeStrategyUpdates(exchangeSimulator, strategyCache);

    }

    /**
     * Periodically updates all strategies in the cache in parallel using the SOSS PMI engine.
     * 
     * @param exchangeSimulator  The simulated feed that provides incoming stock price updates.
     * @param strategyCache      The named cache containing strategy objects that require price updates.
     */
    public static void invokeStrategyUpdates(ExchangeSimulator exchangeSimulator, NamedCache strategyCache)
    {
        // A named cache that will contain a single summary object--this object will contain
        // a summary of alerted strategies. The AlertForm GUI app can look at use this object
        // to get a quick summary of alerted strategies whose positions have fallen outside
        // of desired boundaries.
        NamedCache strategyAlertSummaryCache;
        
        // A named cache containing a single settings object. These settings are shared and available
        // to all workers and UI clients.
        NamedCache settingsCache;
        
        try {
            strategyAlertSummaryCache = CacheFactory.getCache(Settings.SUMMARY_CACHE_NAME);
            settingsCache = CacheFactory.getCache(Settings.SETTINGS_CACHE_NAME);
        } catch (NamedCacheException ex) {
            System.out.println("Unable to get strategy/settings cache: " + ex.getMessage());
            return;
        }
        
        while (true)
        {
            try {
                // The UI app might have changed the settings on us, so we retrieve the settings object
                // prior to every invocation.
                Settings settings = (Settings)settingsCache.get(Settings.SETTINGS_OBJ_NAME);
                if (settings == null)
                {
                    // We're the first client to try to access settings. Add it to the named cache.
                    settings = new Settings();
                    settingsCache.put(Settings.SETTINGS_OBJ_NAME, settings);
                }
                
                // Prepare arguments that will be passed to all workers handling strategy updates:
                StrategyUpdateArgs strategyUpdateArgs = new StrategyUpdateArgs(exchangeSimulator.getPriceSnapshot(), settings.getMaxAllowedDeviationPercent());
                
                // Perform the parallel invocation:
                long startTime = System.nanoTime();
                InvokeResult<AlertSummary> result = strategyCache.invoke(StrategyUpdater.class,
                                                                        null,
                                                                        strategyUpdateArgs,
                                                                        TimeSpan.fromMinutes(1));
                long elapsedTimeMillis = (System.nanoTime() - startTime) / 1000000;
                
                List<InvokeException> errors = result.getErrors();
                if (errors != null && errors.size() > 0)
                {
                     errors.get(0).printStackTrace();
                     return;
                }
                AlertSummary alertSummary = result.getResult();
                
                if (elapsedTimeMillis < MINIMUM_INTERVAL_BETWEEN_PMI_CALLS)
                    alertSummary.setElapsedEvaluationTime(MINIMUM_INTERVAL_BETWEEN_PMI_CALLS);
                else
                    alertSummary.setElapsedEvaluationTime(elapsedTimeMillis);

                // Put the result of the invocation (a summary of alerted strategies) into the 
                // data grid so that the Swing UI (alertsdesk) can access it easily:
                strategyAlertSummaryCache.put(Settings.SUMMARY_OBJ_NAME, alertSummary);
                
                System.out.format("%tT: Invoke completed in %d milliseconds\n",
                                    Calendar.getInstance(), elapsedTimeMillis);
                
                // We update strategy prices once per second, at most. Decide whether we need to sleep.
                if (elapsedTimeMillis < MINIMUM_INTERVAL_BETWEEN_PMI_CALLS)
                {
                    // Prices in the strategies were updated and evaluated with time to spare. Sleep
                    // for the remainder of the 1-second interval.
                    try {
                        Thread.sleep(MINIMUM_INTERVAL_BETWEEN_PMI_CALLS - elapsedTimeMillis);
                    } catch (InterruptedException e) {
                        return;
                    }
                }

            } catch (NamedCacheException ex) {
                System.out.println("Unable to update strategy prices: " + ex.getMessage());
            }
        }
        
    }

    
    /**
     * Generates a collection of stock symbols, along with their initial prices.
     *
     * @param count number of stocks to generate
     * @return a HashMap of Security objects.
     */
    public static HashMap<String, Security> generateInitialSecurities(int count) throws IOException {
        HashMap<String, Security> securities = new HashMap<>(count);
        Random randomPriceGenerator = new Random();

        InputStream inputStream = Main.class.getResourceAsStream("Symbols.txt");
        if (inputStream == null)
            throw new RuntimeException("Symbols.txt not found.");

        try (BufferedReader reader = new BufferedReader(new InputStreamReader(inputStream))) {
            for (int i = 0; i < count; i++) {
                String stockSymbol = reader.readLine();
                if (stockSymbol != null && stockSymbol.length() > 0) {
                    // Give the stock a price between 1 and 100. Note that we avoid a price of zero:
                    BigDecimal price = BigDecimal.valueOf(randomPriceGenerator.nextInt(99) + 1);
                    Security security = new Security(stockSymbol, price, Calendar.getInstance());
                    securities.put(security.getTicker(), security);
                } else {
                    throw new RuntimeException("Not enough stock symbols in Symbols.txt");
                }
            }
        }

        return securities;
    }


    /**
     * Generates simulated strategy objects and inserts them into the ScaleOut in-memory data grid.
     *
     * @param count                number of strategies to create
     * @param positionsPerStrategy number of positions in each strategy. Both a core and hedge position is created for
     *                             the specified value here, so the total positions will be double this parameter.
     * @param securities           collection of stocks (Security objects consisting of ticker symbols and initial
     *                             prices) that will be used in the strategies' positions
     * @return                     list of strategy names that were created and inserted into the in-memory data grid
     */
    public static ArrayList<String> createStrategies(int count, int positionsPerStrategy, HashMap<String, Security> securities) throws NamedCacheException {
        ArrayList<String> strategyNames = new ArrayList<>(count);
        float percentPerPosition = 100f / positionsPerStrategy;
        Random rand = new Random();
        int minStrategyValue = 10000;
        int maxStrategyValue = 1000000;

        NamedCache strategyCache = CacheFactory.getCache(Settings.STRATEGY_CACHE_NAME);
        for (int i = 0; i < count; i++)
        {
            Strategy strategy = new Strategy(i);

            // To initially generate the simulated holdings for a strategy, we'll first come up with a total
            // core/hedge value for the strategy, and then we'll calculate the number of shares we'd need in each
            // position based on the price of the security and and percentage it occupies in the strategy. This
            // is a little backwards from reality, but it lets us easily generate the initial strategies with
            // correct quantities and lets us use security prices that are consistent between strategies.

            int totalCoreVal = rand.nextInt(maxStrategyValue - minStrategyValue + 1) + minStrategyValue;
            int totalHedgeVal = rand.nextInt(maxStrategyValue - minStrategyValue + 1) + minStrategyValue;

            // converting the map of securities to an array in order to ease random selection:
            Security[] secColl = securities.values().toArray(new Security[securities.size()]);

            for (int positionId = 0; positionId < positionsPerStrategy; positionId++)
            {
                Security corePosSecurity = secColl[rand.nextInt(securities.size())];
                Position corePos = new Position.Builder(corePosSecurity.getTicker(), PositionType.Core)
                        .targetAllocationPercent(BigDecimal.valueOf(percentPerPosition))
                        .price(corePosSecurity.getLatestPrice())
                        .shares(BigDecimal.valueOf(
                                totalCoreVal / percentPerPosition / corePosSecurity.getLatestPrice().intValue())
                                .setScale(0, RoundingMode.HALF_UP))
                        .build();
                strategy.getPositions().add(corePos);

                Security hedgePosSecurity = secColl[rand.nextInt(securities.size())];
                Position hedgePos = new Position.Builder(hedgePosSecurity.getTicker(), PositionType.Hedge)
                        .targetAllocationPercent(BigDecimal.valueOf(percentPerPosition))
                        .price(hedgePosSecurity.getLatestPrice())
                        .shares(BigDecimal.valueOf(
                                totalHedgeVal / percentPerPosition / hedgePosSecurity.getLatestPrice().intValue())
                                .setScale(0, RoundingMode.HALF_UP))
                        .build();
                strategy.getPositions().add(hedgePos);
            }

            // Add the strategy to the distributed data grid:
            String strategyName = Integer.toString(strategy.getId());
            strategyCache.insert(strategyName, strategy, strategyCache.getDefaultCreatePolicy(), true, false);

            strategyNames.add(strategyName);
        }

        return strategyNames;
    }
}
